/*
  This file is part of the Linux driver for the Eltron p310c plastic card printer
    Copyright (C) 2002 Flemming Frandsen <ff-p310c@partyticket.net>
    Based on initial work by Stian Skjelstad <stian@nixia.no>
    Materials (printer+cards+ribbons) provided by http://www.psi.no/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/
#include "images.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

void copyImg(Image src, Image dst, int x0, int y0) {
    Pixel *source = src.imagedata;
    Pixel *target = dst.imagedata+y0*dst.width+x0;
    int ll = sizeof(int)*src.width;

    for (int y = 0; y<src.height; ++y) {
	memcpy(target,source,ll);
	target += dst.width;
	source += src.width;
    }
}

Image makeImage(int width, int height) {
    Image img;
    img.height = height;
    img.width = width;
    int bytes = sizeof(int)*img.width*img.height;
    img.imagedata = (unsigned int *)malloc(bytes);    
    memset(img.imagedata,0,bytes);
    return img;
}


CMYKrImage makeCMYKrImage(int width, int height) {
    CMYKrImage cmy;
    cmy.width = width;
    cmy.height = height;
    int size = width*height;
    for (int color=0;color < 4; color++) {
	cmy.color[color] = (unsigned char *)malloc(size);
    }
/*
    cmy.cyan = (unsigned char *)malloc(size);
    cmy.magenta = (unsigned char *)malloc(size);
    cmy.yellow = (unsigned char *)malloc(size);
    cmy.resin_black = (unsigned char *)malloc(size);
*/
    return cmy;
}

CMYKrImage *newCMYKrImage(int width, int height) {
    CMYKrImage *img = (CMYKrImage *)malloc(sizeof(CMYKrImage));
    *img = makeCMYKrImage(width,height);
    return img;
}

CMYKrImage rgb2cmykr(Image rgb, Image *overlay) {

    if (overlay) {
	if (overlay->width != rgb.width ||
	    overlay->height != rgb.height) {
	    fprintf(stderr,"Error: The overlay image must be the same size as the background (%d,%d) != (%d,%d)\n",
		    overlay->width, overlay->height,
		    rgb.width, rgb.height);
	    exit(-4);
	}
    }

    // This is the number of pixels / bits pr cmy part:
    int size = rgb.width*rgb.height*sizeof(unsigned char);
    
    CMYKrImage cmy = makeCMYKrImage(rgb.width,rgb.height);
    for (int pixel = 0; pixel < size; pixel++) {
	// Start out by letting the bottom / background set the output color:
	unsigned int cyan    = 0;
	unsigned int magenta = 0;
	unsigned int yellow  = 0;
	unsigned char resin  = 0;

	unsigned int rgbpixel = rgb.imagedata[pixel];
	if (rgbpixel & 0x00ffffff) {
	    cyan    = 255-((rgbpixel >> 16) & 0xff);
	    magenta = 255-((rgbpixel >> 8)  & 0xff);
	    yellow  = 255-( rgbpixel        & 0xff);
	    resin = 0;

//	    fprintf(stderr,"pixel: %d; rgb:%06x; cmyk: %d,%d,%d,%d\n",
//		    pixel, rgbpixel, cyan,magenta,yellow,resin);

	} else { 
	    // All black so just put the resin there 
	    // as this is the bottom layer also turn up the CMY to max or we will get alignment issues...
	    cyan = 255;
	    magenta = 255;
	    yellow = 255;
	    resin = 1;
	}
	
	// Then layer the overlay image ontop of that:
	if (overlay) {
	    unsigned int opixel = overlay->imagedata[pixel];
	    int alpha = opixel >> 24;
	    if (!alpha) {
		// Don't do anything to this pixel at all...
	    } else if ((opixel & 0x00ffffff) || alpha != 255) {
		cyan    = (cyan    * (255-alpha) + alpha * (255-(((opixel >> 16) & 0xff)))) / 255;
		magenta = (magenta * (255-alpha) + alpha * (255-(((opixel >> 8)  & 0xff)))) / 255;
		yellow  = (yellow  * (255-alpha) + alpha * (255-(( opixel        & 0xff)))) / 255;
		resin = 0; // Dont put resin here if that's what the bottom layer wanted...
	    } else {
		resin = 1; // Just turn on the resin, but don't touch the CMY from the bottom layer
	    }
	}
/*
	cmy.cyan[pixel]        = cyan >> 3;
	cmy.magenta[pixel]     = magenta >> 3;
	cmy.yellow[pixel]      = yellow >> 3;
	cmy.resin_black[pixel] = resin;
*/
	cmy.color[CYAN][pixel]    = cyan >> 3;
	cmy.color[MAGENTA][pixel] = magenta >> 3;
	cmy.color[YELLOW][pixel]  = yellow >> 3;
	cmy.color[BLACK][pixel]   = resin;
    }

    return cmy;
}

void destroyImage(Image img) {
    free(img.imagedata); img.imagedata = NULL;
}

void destroyCMYKrImage(CMYKrImage cmy) {
    for (int color=0;color < 4; color++) {
	free(cmy.color[color]);
	cmy.color[color] = NULL;
    }
/*    
    free(cmy.cyan);        cmy.cyan = NULL;
  p  free(cmy.magenta);     cmy.magenta = NULL;
    free(cmy.yellow);      cmy.yellow = NULL;
    free(cmy.resin_black); cmy.resin_black = NULL;
*/
}

const int MAGIC = 0x0F155E + 1;

CMYKrImage *loadCMYKrImage(char *fn) {
    FILE *f = fopen(fn, "r");
    if (f) {
	int magic = 0;
	int height = 0;
	int width = 0;
	if (fread(&magic,sizeof(magic),1,f) == 1 &&
	    magic == MAGIC &&
	    fread(&width,sizeof(width),1,f) == 1 &&
	    fread(&height,sizeof(height),1,f) == 1 &&
	    width > 0 && height > 0 && width * height < 1000000) {

	    CMYKrImage *img = newCMYKrImage(width, height);
	    for (int color=0;color < 4; color++) {
		fread(img->color[color],width*height,1,f);
	    }
/*
	    fread(img->cyan,width*height,1,f);
	    fread(img->magenta,width*height,1,f);
	    fread(img->yellow,width*height,1,f);
	    fread(img->resin_black,width*height,1,f);
*/
	    fclose(f);
	    return img;
	} else {
	    fclose(f);
	}
    }
    return NULL;
}

bool saveCMYKrImage(char *fn, CMYKrImage *img) {
    FILE *f = fopen(fn, "w");
    if (f) {
	int magic = MAGIC;
	fwrite(&magic,sizeof(magic),1,f);
	fwrite(&img->width,sizeof(img->width),1,f);
	fwrite(&img->height,sizeof(img->height),1,f);
	for (int color=0;color < 4; color++) {
	    fwrite(img->color[color], img->width*img->height,1,f);
	}
/*	
	fwrite(img->cyan, img->width*img->height,1,f);
	fwrite(img->magenta,img->width*img->height,1,f);
	fwrite(img->yellow,img->width*img->height,1,f);
	fwrite(img->resin_black,img->width*img->height,1,f);
*/
	fclose(f);
	return true;
    } else {
	return false;
    }
}

char *monochrome(CMYKrImage img, int x, int y, int width, int height) {
    if (width % 8) {
	fprintf(stderr, "width of monochrome area is not divisable by 8: %i\n",width);
	exit(-11);
    }
    if (x % 8) {
	fprintf(stderr, "x location of monochrome area is not divisable by 8: %i\n",x);
	exit(-11);
    }

    char *res = (char *)malloc(width*height/8);    
    
    int j = 0;
    int i = 0;
    for (int yc = 0; yc<height; yc++) {
	int bit = 0;
	char byte = 0;
	for (int xc = 0; xc<width; xc++) {
	    if (img.color[BLACK][x+xc+img.width*(y+yc)]) {
		byte |= 1 << (7-bit);
	    }	
	    if (++bit > 7) {
		bit = 0;
		res[j++] = byte;
		byte = 0;
	    }	    
	}
    }
    return res;
}

char *colorrect(CMYKrImage img, int x, int y, int width, int height, int color) {
    unsigned char *pixels = NULL;

    if (color < 0 || color > 3) {
	fprintf(stderr,"Invalid color passed to colorrect: %d\n",color); 
	exit(-1);
    }

    char *res = (char *)malloc(width*height);
    int j = 0;
    for (int yc = 0; yc<height; yc++) {
	for (int xc = 0; xc<width; xc++) {
	    res[j++] = img.color[color][(yc+y)*img.width+(xc+x)];
	}
    }
    return res;
}
